/*

*/

package language1;

import tree.*;
import tokenizer.*;

/**
   Convert an Abstract Syntax Tree (AST) for the language
   of expressions into an infix expression.
   <p>
   The ast2infix() method takes an Abstract Syntax Tree
   for the language of arithmetic and boolean expressions
   and converts it into a string that contains the infix
   version of the expression.
   <p>
   ast2infix() uses knowledge about operator precedence
   (in infix notation) to use the minimum number of
   parentheses.
*/
public class AST2infix
{
   /**
      This ast2infix() method is essentially
      a in-order traversal of the tree.

      @param tree  an expression {@link Tree} to be converted to infix notation
      @return the {@link String} representation of the expression {@link Tree}'s infix notation
   */
   public static String ast2infix(final Tree<Token> tree)
   {
      String result = "";

      if ( tree.subTrees().size() == 0 )  // constant term
      {
         result += tree.root().lexeme;
      }
      else if ( tree.subTrees().size() == 1 )  // unary operator
      {
         if ( tree.subTrees().get(0).subTrees().size() >= 2 )
         {
            result += tree.root().lexeme + "(" + ast2infix( tree.subTrees().get(0) ) + ")";
         }
         else
         {
            result += tree.root().lexeme + ast2infix( tree.subTrees().get(0) );
         }
      }
      else if ( tree.subTrees().size() == 2 )  // binary operator
      {
         final int middlePrecedence = precedence( tree.root().lexeme );
         final int leftPrecedence = precedence( tree.subTrees().get(0).root().lexeme );

         if ( leftPrecedence > middlePrecedence )
         {
            result += "(" + ast2infix( tree.subTrees().get(0) ) + ")";
         }
         else
         {
            result += ast2infix( tree.subTrees().get(0) );
         }

         result += " " + tree.root().lexeme + " ";

         final int rightPrecedence = precedence( tree.subTrees().get(1).root().lexeme );

         if ( rightPrecedence >= middlePrecedence )
         {
            result += "(" + ast2infix( tree.subTrees().get(1) ) + ")";
         }
         else
         {
            result += ast2infix( tree.subTrees().get(1) );
         }
      }
      else  // operators with 3 or more operands
      {
         final int operatorPrecedence = precedence( tree.root().lexeme );
         final int leftPrecedence = precedence( tree.subTrees().get(0).root().lexeme );

         if ( leftPrecedence > operatorPrecedence )
         {
            result += "(" + ast2infix( tree.subTrees().get(0) ) + ")";
         }
         else
         {
            result += ast2infix( tree.subTrees().get(0) );
         }

         for (int i = 1; i < tree.degree(); ++i)
         {
            result += " " + tree.root().lexeme + " ";

            final int rightPrecedence = precedence( tree.subTrees().get(i).root().lexeme );

            if ( rightPrecedence >= operatorPrecedence )
            {
               result += "(" + ast2infix( tree.subTrees().get(i) ) + ")";
            }
            else
            {
               result += ast2infix( tree.subTrees().get(i) );
            }
         }
      }
      return result;
   }


   // http://www.java-tips.org/java-se-tips/java.lang/what-is-java-operator-precedence.html
   // http://www.cs.princeton.edu/introcs/11precedence/
   private static int precedence(final String op)
   {
      int result = 0;  // "highest" precedence (i.e., constants)

      if ( op.equals("!") )
      {
         result = 2;
      }
      else if ( op.equals("**") )
      {
         result = 3;
      }
      else if ( op.equals("*")
             || op.equals("/")
             || op.equals("%") )
      {
         result = 4;
      }
      else if ( op.equals("+")
             || op.equals("-") )
      {
         result = 5;
      }
      else if ( op.equals("<")
             || op.equals(">")
             || op.equals("<=")
             || op.equals(">=") )
      {
         result = 7;
      }
      else if ( op.equals("==")
             || op.equals("!=") )
      {
         result = 8;
      }
      else if ( op.equals("&&") )
      {
         result = 11;
      }
      else if ( op.equals("||") )
      {
         result = 12;
      }

      return result;
   }
}
