/*

*/

package language1;

import tree.*;
import tokenizer.*;

/**
   Evaluate (interpret) an Abstract Syntax Tree (AST)
   that is defined by the following grammar.
   <pre>{@code
        Expr ::= BExpr
               | AExpr

      BExpr ::= '(' '||'  BExpr BExpr+ ')'  // two or more parameters
              | '(' '&&'  BExpr BExpr+ ')'  // two or more parameters
              | '(' '!'   BExpr ')'         // one parameter
              | '('  EqOp Expr Expr ')'     // two parameters
              | '(' RelOp AExpr AExpr ')'   // two parameters
              | BOOLEAN

       EqOp ::= '==' | '!='
      RelOp ::= '<' | '>' | '<=' | '>='

      AExpr ::= '(' '+' AExpr AExpr* ')'    // one or more parameters
              | '(' '-' AExpr AExpr? ')'    // one or two parameters
              | '(' '*' AExpr AExpr+ ')'    // two or more parameters
              | '(' '/' AExpr AExpr  ')'    // two parameters
              | '(' '%' AExpr AExpr  ')'    // two parameters
              | '(' '**' AExpr AExpr ')'    // two parameters
              | INTEGER

    BOOLEAN ::= 'true' | 'false'
    INTEGER ::= [-|+][0-9]+
   }</pre>

   The '/', ' %', '**' and relational operators operators are strictly binary operators.
   <p>
   The '+' operator can take any number of arguments (i.e., one or more).
   <p>
   The '-' operator can take one or two arguments.
   <p>
   The '*', '{@code &&}', and '||' operators can take two or more arguments.
   <p>
   The '!" operator takes just one argument.
   <p>
   The boolean '{@code &&}' and '||' operators should be short circuited.
   <p>
   The equality operators, '==' and '!=', work with both data types, integer and boolean.
   <p>
   The relational operators only work with the integer data type.
   <p>
   The result of evaluating an Expr is either an integer value or a boolean value.
   <p>
   This language is a subset of our language of trees (every string from this
   language is also a string from the tree language.
*/
public class EvaluateAST
{
   /**
      The methods eval(), evaluateA(), evaluateB(), and evaluateR() are
      essentially a post-order traversal of the abstract syntax tree.

      @param tree  an expression {@link Tree} to be evaluated
      @return the value of the expression {@link Tree}
   */
   public static Value eval(final Tree<Token> tree)
   {
      final Value result;  // a blank final

      final String node = tree.root().lexeme;

      if ( node.equals("&&")
        || node.equals("||")
        || node.equals("!") )
      {
         result = evaluateBexpr(tree);  // boolean operator
      }
      else if ( node.equals("==")
             || node.equals("!=") )
      {
         result = evaluateEqexpr(tree);  // equality operator
      }
      else if ( node.equals("<")
             || node.equals(">")
             || node.equals("<=")
             || node.equals(">=") )
      {
         result = evaluateRelexpr(tree);  // relational operator
      }
      else if ( node.equals("+")
             || node.equals("-")
             || node.equals("*")
             || node.equals("/")
             || node.equals("%")
             || node.equals("**") )
      {
         result = evaluateAexpr(tree);  // arithmetic operator
      }
      else if ( tree.subTrees().size() == 0 )  // constant term
      {
         if ( node.equals("true") || node.equals("false") )
         {
            result = new Value( node.equals("true") );
         }
         else if ( node.matches("^[-|+]*[0-9][0-9]*") )
         {
            result = new Value( Integer.parseInt( node ) );
         }
         else  // runtime check
         {
            throw new EvalError("undefined constant: " + node);
         }
      }
      else
      {
         throw new EvalError("invalid expression: "
                             + toString(tree));
      }

      return result;
   }


   // Evaluate a boolean expression.
   private static Value evaluateBexpr(final Tree<Token> tree)
   {
      if ( 0 == tree.subTrees().size() )  // runtime check
      {
         throw new EvalError("missing arguments: "
                             + toString(tree));
      }

      boolean result = false;

      final String node = tree.root().lexeme;

      Value value = eval( tree.subTrees().get(0) );
      if ( ! value.tag.equals(Value.BOOL_TAG) )  // runtime check
      {
         throw new EvalError("not a boolean expression: "
                              + toString(tree.subTrees().get(0)));
      }
      result = value.valueB;

      if ( node.equals("&&") )
      {
         if ( 2 > tree.subTrees().size() )  // runtime check
         {
            throw new EvalError("wrong number of arguments: "
                                + toString(tree));
         }

         for (int i = 1; i < tree.subTrees().size(); i++)
         {
            if (result)
            {
               value = eval( tree.subTrees().get(i) );
               if ( ! value.tag.equals(Value.BOOL_TAG) )  // runtime check
               {
                  throw new EvalError("not a boolean expression: "
                                      + toString(tree.subTrees().get(i)));
               }
               result = result && value.valueB;
            }
            else  // short circuit the evaluation of '&&'
            {
               result = false;
               break;
            }
         }
      }
      else if ( node.equals("||") )
      {
         if ( 2 > tree.subTrees().size() )  // runtime check
         {
            throw new EvalError("wrong number of arguments: "
                                + toString(tree));
         }

         for (int i = 1; i < tree.subTrees().size(); i++)
         {
            if (! result)
            {
               value = eval( tree.subTrees().get(i) );
               if ( ! value.tag.equals(Value.BOOL_TAG) )  // runtime check
               {
                  throw new EvalError("not a boolean expression: "
                                      + toString(tree.subTrees().get(i)));
               }
               result = result || value.valueB;
            }
            else  // short circuit the evaluation of '||'
            {
               result = true;
               break;
            }
         }
      }
      else if ( node.equals("!") )
      {
         if ( 1 != tree.subTrees().size() )  // runtime check
         {
            throw new EvalError("wrong number of arguments: "
                                + toString(tree));
         }
         result = ! result;
      }

      return new Value( result );
   }//evaluateBexpr()


   // Evaluate an equality expression (which is a boolean expression).
   private static Value evaluateEqexpr(final Tree<Token> tree)
   {
      if ( 2 != tree.subTrees().size() )  // runtime check
      {
         throw new EvalError("wrong number of arguments: "
                             + toString(tree));
      }

      boolean result = false;

      final String op = tree.root().lexeme;

      final Value valueL = eval( tree.subTrees().get(0) );
      final Value valueR = eval( tree.subTrees().get(1) );

      if ( op.equals("==") )
      {
         if ( ! valueL.tag.equals(valueR.tag) )
         {
            result = false;
         }
         else if ( valueL.tag.equals(Value.INT_TAG) )
         {
            final int resultL = valueL.valueI;
            final int resultR = valueR.valueI;
            result = resultL == resultR;
         }
         else // boolean data type
         {
            final boolean resultL = valueL.valueB;
            final boolean resultR = valueR.valueB;
            result = resultL == resultR;
         }
      }
      else // the '!=' operator
      {
         if ( ! valueL.tag.equals(valueR.tag) )
         {
            result = true;
         }
         else if ( valueL.tag.equals(Value.INT_TAG) )
         {
            final int resultL = valueL.valueI;
            final int resultR = valueR.valueI;
            result = resultL != resultR;
         }
         else // boolean data type
         {
            final boolean resultL = valueL.valueB;
            final boolean resultR = valueR.valueB;
            result = resultL != resultR;
         }
      }

      return new Value( result );
   }//evaluateEqexpr()


   // Evaluate a relational expression (which is a boolean expression).
   private static Value evaluateRelexpr(final Tree<Token> tree)
   {
      if ( 2 != tree.subTrees().size() )  // runtime check
      {
         throw new EvalError("wrong number of arguments: "
                             + toString(tree));
      }

      boolean result = false;

      final String op = tree.root().lexeme;

      final Value valueL = eval( tree.subTrees().get(0) );
      if ( ! valueL.tag.equals(Value.INT_TAG) )  // runtime check
      {
         throw new EvalError("not an integer expression: "
                             + toString(tree.subTrees().get(0)));
      }

      final Value valueR = eval( tree.subTrees().get(1) );
      if ( ! valueR.tag.equals(Value.INT_TAG) )  // runtime check
      {
         throw new EvalError("not an integer expression: "
                             + toString(tree.subTrees().get(1)));
      }

      final int resultL = valueL.valueI;
      final int resultR = valueR.valueI;

      if ( op.equals("<") )
      {
         result = resultL < resultR;
      }
      else if ( op.equals(">") )
      {
         result = resultL > resultR;
      }
      else if ( op.equals("<=") )
      {
         result = resultL <= resultR;
      }
      else if ( op.equals(">=") )
      {
         result = resultL >= resultR;
      }

      return new Value( result );
   }//evaluateRelexpr()


   // Evaluate an arithmetic expression.
   private static Value evaluateAexpr(final Tree<Token> tree)
   {
      if ( 0 == tree.subTrees().size() )  // runtime check
      {
         throw new EvalError("missing arguments: "
                             + toString(tree));
      }

      int result = 0;

      final String node = tree.root().lexeme;

      final Value valueL = eval( tree.subTrees().get(0) );
      if ( ! valueL.tag.equals(Value.INT_TAG) )  // runtime check
      {
         throw new EvalError("not an integer expression: "
                             + toString(tree.subTrees().get(0)));
      }
      final int resultL = valueL.valueI;
            int resultR = 0;

      if ( tree.subTrees().size() >= 2 )
      {
         final Value valueR = eval( tree.subTrees().get(1) );
         if ( ! valueR.tag.equals(Value.INT_TAG) )  // runtime check
         {
            throw new EvalError("not an integer expression: "
                                + toString(tree.subTrees().get(1)));
         }
         resultR = valueR.valueI;
      }

      if ( node.equals("+") )
      {
         if ( tree.subTrees().size() == 1 )
            result = resultL;
         else
         {
            result = resultL + resultR;

            for (int i = 2; i < tree.subTrees().size(); i++)
            {
               final Value temp = eval( tree.subTrees().get(i) );
               if ( ! temp.tag.equals(Value.INT_TAG) )  // runtime check
               {
                  throw new EvalError("not an integer expression: "
                                      + toString(tree.subTrees().get(i)));
               }
               result += temp.valueI;
            }
         }
      }
      else if ( node.equals("-") )
      {
         if ( 2 < tree.subTrees().size() )  // runtime check
         {
            throw new EvalError("wrong number of arguments: "
                                + toString(tree));
         }
         if ( tree.subTrees().size() == 1 )
            result = -resultL;
         else
            result = resultL - resultR;
      }
      else if ( node.equals("*") )
      {
         if ( 1 == tree.subTrees().size() )  // runtime check
         {
            throw new EvalError("wrong number of arguments: "
                                + toString(tree));
         }

         result = resultL * resultR;

         for (int i = 2; i < tree.subTrees().size(); i++)
         {
            final Value temp = eval( tree.subTrees().get(i) );
            if ( ! temp.tag.equals(Value.INT_TAG) )  // runtime check
            {
               throw new EvalError("not an integer expression: "
                                    + toString(tree.subTrees().get(i)));
            }
            result *= temp.valueI;
         }
      }
      else if ( node.equals("/") )
      {
         if ( 2 != tree.subTrees().size() )  // runtime check
         {
            throw new EvalError("wrong number of arguments: "
                                + toString(tree));
         }
         result = resultL / resultR;
      }
      else if ( node.equals("%") )
      {
         if ( 2 != tree.subTrees().size() )  // runtime check
         {
            throw new EvalError("wrong number of arguments: "
                                + toString(tree));
         }
         result = resultL % resultR;
      }
      else if ( node.equals("**") )
      {
         if ( 2 != tree.subTrees().size() )  // runtime check
         {
            throw new EvalError("wrong number of arguments: "
                                + toString(tree));
         }
         result = (int)Math.pow(resultL, resultR);
      }

      return new Value( result );
   }//evaluateAexpr()


   /**
      Convert a {@code Tree} of {@link Token}s into a {@link String}.
      <p>
      This method does a pre-order traversal of the {@link Tree}.
      <p>
      This method is used by the evaluation errors.

      @param tree  {@link Tree} of {@link Token}s to convert into a {@link String}
      @return a {@link String} representation of this {@code Tree} of {@link Token}s
   */
   public static String toString(final Tree<Token> tree)
   {
      String result = "";
      if ( tree.isLeaf() )
      {
         result += tree.root().lexeme;
      }
      else
      {
         result += "(" + tree.root().lexeme;
         for ( final Tree<Token> subTree : tree.subTrees() )
         {
            result += " " + toString(subTree);  // recursion
         }
         result += ")";
      }
      return result;
   }
}
